
//	The Hacker's Diet

/*	Define NDEBUG for production builds.  All assertions and other
	in-line debug code will be removed.  */

#define NDEBUG

#define NON_INTERNATIONAL			// Permit access to IsDigit(), etc. on Palm OS 3.1 SDK

#include <PalmOS.h>
#include <CharAttr.h>

#include "HackdietRsc.h"
#ifdef __GNUC__
#include "Callbacks.h"
#endif

#define OS20BUG_WORKAROUND			//	Include workarounds for bugs in PalmOS 2.0 ROM		

#ifndef NDEBUG
#ifdef ERROR_CHECK_LEVEL
#undef ERROR_CHECK_LEVEL
#endif
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#define assert(x) ErrFatalDisplayIf(!(x), "Assertion failed: " # x)
#else
#ifdef ERROR_CHECK_LEVEL
#undef ERROR_CHECK_LEVEL
#endif
#define ERROR_CHECK_LEVEL ERROR_CHECK_NONE
#define assert(x)
#endif

#define CSV							// Support CSV import
//#define CSVDEBUG					// Include CSV debugging support
#define BACKUPDB					// Set backup bit on monthly log database
//#define FORCEUNITS				// Force units from preferences into old database

/*	If you uncomment the following line to define FIND_SUPPORT, code
	will be enabled which responds to the global find mechanism in
	PalmOS, searching and flagging months which contain the target
	string in one or more comment fields.  When these links are
	tapped, the log for that month is shown with the day whose
	comment matched scrolled into view.	
*/

#define FIND_SUPPORT				// Support global find searching comments ?

#define CreatorID	'HkDt'			// Application Creator ID
#define VersionNo	100				// Version number * 100 (215 == ver 2.15)

//	Database identities

#define MonthDbType	'Mlog'			// Monthly log database type
#define MonthDbName	"Months-HkDt"	// Monthly log database name
#ifdef CSV
#define ImportDbType 'Wcsv'			// CSV import database type
#endif

//	Extrema

#define MAX_MONTH_NAME	20			// Bytes (consider DBCS) in longest month name
#define MAX_MONTH_YYYY	(MAX_MONTH_NAME + 6)  // Max bytes in "Month, YYYY"
#define MAX_COMMENT		80			// Bytes (less trailing 0) in longest comment

//	Type definitions

typedef unsigned short weight;		// Weight (integer with 2 fixed decimal places)
typedef unsigned long trend;		// Trend (long with 6 fixed decimal places)
typedef unsigned char rung;			// Exercise rung (8 bit integer, 255 = no entry)

#define WEIGHT_MAX	0xFFFF
#define TREND_MAX	0xFFFFFFFF
#define RUNG_MAX	48
#define RUNG_ABSENT	255

#define WeightToTrend(w) ((w) * 10000L)
#define TrendToWeight(t) (((t) + 5000L) / 10000L)

/*	When the weight unit is set to stones, we still actually keep the log
	and do computations using pounds.  Consequently, the following
	definition distinguishes only between kilograms and other units.  */

#define CaloriesPerWeightUnit(wu)	(((wu) == wKilogram) ? 7716 : 3500)

/*	Calories (gram) per Joule, and hence Calories (kilogram)
	per kiloJoule.  */

#define CaloriesPerJoule			0.239045

//	Selected energy units per weight unit

#define EnergyPerWeightUnit(wu)	(CaloriesPerWeightUnit(wu) / (PrefFlag(pfEnergyJoule) ? \
									CaloriesPerJoule : 1))			

#define min(x, y) (((x) < (y)) ? (x) : (y))
#define max(x, y) (((x) > (y)) ? (x) : (y))
#define abs(x) (((x) < 0) ? (-(x)) : (x))

/* The following definitions create non-absurdly-named macros
   which we can use to invoke the absurd TxtGlue... functions
   to work around Palm's strategic incompatibility in handling of
   international character sets.  You see, first there were calls
   like IsAlpha(GetCharAttr(), ch), but they worked only for Latin
   characters.  So then, they added the "International Feature Set"
   to handle other character sets, with its own plethora of calls
   like TxtCharIsAlpha(ch).  Fine, but of course if you try to
   access one of these calls on an older ROM which doesn't have
   the International Feature Set, blooie...it's paper clip time.
   
   Applications which don't support or require anything beyond
   the Latin character set could, of course, continue to use the
   old GetCharAttr() calls first, without source code changes and
   then, as Palm turned the screw tighter, by the expedient of
   defining NON_INTERNATIONAL before including PalmOS.h.  This
   brings us to the 4.0 debug ROM, where any call at all on
   GetCharAttr() results in the absolutely comprehensible "Fatal
   Error", "TextTable.cpp, Line: 1871, Incorrect text table version"
   which, after several hours of trial, error, and hair pulling you
   discover really means, "We don't want you to call those old-fashioned
   GetCharAttr() functions any more, even though they continue to work
   perfectly on the production ROMs."
   
   One with a more conspiratorial view of the world than this humble
   programmer might impute to Palm Microsoft-like motives of encouraging
   application developers to use the new TxtChar... calls, which would
   cause their applications not to work on older handhelds, stimulating
   sales on the hardware side of the shop.  But no....  Palm taketh away,
   and Palm giveth, kind of.  They've introduced the so-called "PalmOSGlue
   Library" which provides a set of functions which use the new calls if
   they're present and the old ones otherwise.  But this means what was
   once in the days of C a macro like isalpha(ch) now bloats into
   TxtGlueCharIsAlpha(ch) which does heaven knows what kind of
   conditional tests inside your time-critical character fetch loop.
   And, oh yes, this is a *link* library, so using it inflates the
   size of your application.
   
   Fortunately, this program does very little bulk character parsing.
   The only notable case is in CSV import, and that's something most
   users will only ever do once.  So we define palatable variants for
   the hideous Glue functions below and use them throughout.  That way,
   we'll only have the change the definitions here the next time the
   Palmies change their minds.
   
   One more little wrinkle...if you should be so imprudent as to
   pass an ISO character with a character code > 127 to one of these
   TxtChar... functions, it'll blow you off on the debug ROM with a
   "Fatal Error" because the 16-bit value it got was sign extended.
   So we wrap the additional crapola toAscii() call around the
   argument to make sure that doesn't happen.
   
*/

#include <TxtGlue.h>

#define toAscii(c) ((c) & 0xFF)
#define toLower(c) TxtGlueLowerChar(toAscii(c))
#define toUpper(c) TxtGlueUpperChar(toAscii(c))

#define isAlpha(c) TxtGlueCharIsAlpha(toAscii(c))
#define isDigit(c) TxtGlueCharIsDigit(toAscii(c))
#define isGraph(c) TxtGlueCharIsGraph(toAscii(c))
#define isLower(c) TxtGlueCharIsLower(toAscii(c))
#define isPrint(c) TxtGlueCharIsPrint(toAscii(c))
#define isSpace(c) TxtGlueCharIsSpace(toAscii(c))
#define isUpper(c) TxtGlueCharIsUpper(toAscii(c))

/*	We also define a few character class macros of our own in
    the same style for tests Palm doesn't support or which don't
    have equivalents (for example IsAscii) in the "Glue" library. */

#define isAscii(c)	(((c) >= 0) && ((c) <= 255))
#define isOctal(c)	(((c) >= '0') && ((c) <= '7'))

//	Standard palette colour intensities

#define	CI_0	0x00
#define CI_1	0x33
#define CI_2	0x66
#define CI_3	0x99
#define CI_4	0xCC
#define CI_5	0xFF
#define setRGBColour(cs, ri, gi, bi)	cs.index = 0, cs.r = (ri), cs.g = (gi), cs.b = (bi)

//	Preferences and permanent globals record

typedef enum { wKilogram = 0, wPound = 1, wStone = 2 } weightunit;

//	Flag bits for preFlags

#define pfPlotPlan		0			// Plot plan in chart
#define pfNoColour		1			// Suppress colour on colour device
#define pfEnergyJoule	2			// Show energy in kJ instead of kcal

struct hdPreferences {
	unsigned char monthsKeep;		// Months to keep on handheld
	unsigned short monthsKept;		// Months actually present in database
	weightunit weightUnit;			// Unit used in weight logs
	Boolean privateLogs;			// True if log data are deemed private
	Boolean protectEntries;			// Protect entries in monthly log
	weightunit displayUnit;			// Unit used to display weight
	rung currentRung;				// Current rung in exercise ladder
	unsigned short preFlags;		// Boolean flag bits
	
	//	The following variables save the peristent state of the diet calculator
	
	int calcCalorieBalance;			// Calorie balance
	weight calcStartWeight;			// Start weight
	weight calcGoalWeight;			// Weight goal
	UInt32 calcStartDate;			// Start date
	
	/*	The following structure defines the colour scheme used by
		various forms when running on colour platform and the flat
		pfNoColour has not been set to suppress colour.  Each entry
		gives the RGB components of the colour (which should conform
		to values present in the standard palette).  The index field
		in the RGBColorType is filled in with the slot containing
		this colour (or the closest match) after the preferences
		have been loaded, and the index is subsequently used when
		drawing with the colour.  */
	
	struct {						// Colours used in various forms
		RGBColorType
			logPlusVarianceColour,	// Monthly log positive variances
			logMinusVarianceColour,	// Monthly log negative variances
			
			chartBackgroundColour,	// Chart background colour
			chartTrendColour,		// 		Trend colour index
			chartFlsinkColour,		// 		Float and sinker line colour index
			chartBobColour,			// 		Bob outline colour
			chartFlagColour,		// 		Flagged day colour
			chartRungColour,		// 		Exercise rung colour
			chartPlanColour;		// 		Diet plan colour


	} colourScheme;
};

//	Monthly log record

struct hdMonthlyLog {
	unsigned short year;			// Year for this log (watch not; not Y65K safe!)
	unsigned char month;			// Month (0 - 11)
	unsigned char weightUnit;		// Weight unit used in this record
	trend trendCarryForward;		// Trend carry-forward from last month
	UInt32 lastModificationTime;	// Time (TimGetSeconds) of last modification
	unsigned char flagsArray[4];	// Bit array for maximum of 31 flags
	rung rungArray[31];				// Exercise rung array
	unsigned char padByte1;			// FILLER SIZE IS 1
	weight weightArray[31];			// Weight array
	
	/*	Comments are stored as a running string, terminated by
		a zero byte.  Each comment consists of the following
		sequence of bytes:
		
				0:		Day of month (1 - 31) comment attached to
				1:		Length of comment string
				2-n:	Comment text
				
		The minimum comment array length is one byte, the zero
		terminator.
		
		Duplicate comment strings may (but need not be compressed).
		A compressed comment string is always a two-byte sequence:
		
				0:		Day of month (1 - 31) comment attached to + 32
				1:		Previous day of month's (1 - 31) comment which is identical
	*/
	
	char commentArray[2];			// Comment array
};

//	Unpacked data for month

struct hdMonth {
	unsigned short year;			// Year for this log (watch not; not Y65K safe!)
	unsigned char month;			// Month (0 - 11)
	unsigned char weightUnit;		// Weight unit used in this record
	unsigned int days;				// Days in this month
	unsigned int firstDay;			// Day of week of first day
	trend trendCarryForward;		// Trend carry-forward from last month
	UInt32 lastModificationTime;		// Time (TimGetSeconds) of last modification
	unsigned int dataLength;		// Length of total unpacked data
	//	Note that these arrays are indexed by day numbers 0-30, not 1-31
	rung rungArray[31];				// Exercise rung array
	weight weightArray[31];			// Weight array
	trend trendArray[31];			// Running trend array
	unsigned char flagsArray[31];	// Flags
	short commentArray[31];			/* Relative pointers to comments.  Note that
									   these pointers are relative to the start of
									   the hdMonth structure itself (which must be cast
									   to a char * before adding the commentArray offset),
									   *not* the commentData array!  This rather odd
									   convention allows us to use zero in commentArray
									   to denote the absence of a comment.  Note also that
									   a short is completely adequate for these offsets since
									   comments are limited to 80 characters and there
									   can't be more than 31 comments per month, and
									   hence a maximum offset of only 2480 bytes plus
									   the bytes in the hdMonth structure which precedes
									   the commentArray.  */
	char commentData[2];			// String data for comments
};

//	Function prototypes

//	About.c

extern void aboutDialogue(void);

//	Charting.c

extern void chartingLoad(FormPtr f);

//	Csv.c

#ifdef CSV
extern void CSVscanInit(char *s);
extern Boolean CSVscanField(char *f);
extern void csvRecordInit(char *r, int l);
extern Boolean csvRecordNext(char *dest, int destl);
extern UInt16 csvRecordCount(void);
extern void csvImportLoad(FormPtr f);

//	Csvdebug.c

#ifdef CSVDEBUG
extern void csvDebugLoad(FormPtr f);
#endif

//	Csvin.c

extern Boolean parseDay(char *csvLine,
				 int *year, int *month, int *day,
				 weight *wgt,
				 rung *rng,
				 Boolean *flag,
				 char *comment);
extern void importCSVstart(void);
extern Boolean importCSVline(char *l);
extern void importCSVend(void);				
extern Boolean checkImportCSV(void);
#endif

//	Database.c

extern int hdFoundIndex;				// Database	index of last record found by various functions
extern void updateTrend(struct hdMonth *m);
extern struct hdMonth *unpackMonthRecord(struct hdMonthlyLog *md);
extern struct hdMonth *hdAddComment(struct hdMonth *m, int day, char *comment);
extern UInt16 hdNumMonths(void);
extern void hdIterateMonths(int *iter, UInt16 *year, UInt16 *month);
extern int hdFindMonth(UInt16 dyear, UInt16 dmonth);
extern struct hdMonth *hdGetMonthUnpacked(UInt16 recnr);
extern struct hdMonth *hdGetMonthByDate(UInt16 year, UInt16 month, UInt16 *recnr, Boolean create);
extern Boolean hdPutMonthUnpacked(struct hdMonth *m, UInt16 recnr);
extern void hdDisposeMonthUnpacked(struct hdMonth *m);
extern Boolean hdSaveMonthUnpacked(struct hdMonth *m);
extern Boolean hdIsMonthPresent(UInt16 dyear, UInt16 dmonth);
extern void hdMonthsPresent(int year, Boolean monthsPresent[12], int *previousYear, int *nextYear);
extern void hdEnsureMonthPresent(UInt16 dyear, UInt16 dmonth);
extern Boolean hdEnsureDatabasePresent(void);
extern UInt32 hdFindInComments(struct hdMonthlyLog *md, char * target, UInt16 *pos);
extern void hdWeightExtrema(UInt16 syear, UInt16 smonth, UInt16 eyear, UInt16 emonth,
							weight *weightMin, weight *weightMax);
extern void hdPropagateTrend(UInt16 year, UInt16 month);

//	Dietcalc.c

extern void dietCalcLoad(FormPtr frm);

//	Exercise.c

extern void exerciseLoad(FormPtr f);

//	Find.c

extern void findQuery(FindParamsPtr findParams);
extern void findGoTo(GoToParamsPtr goToParams, Boolean launchingApp);

//	Hackdiet.c

//	Feature set Booleans set based on ROM version
extern Boolean featureSet2_0;			// 2.0 ROM or later
extern Boolean featureSet3_0;			// 3.0 ROM or later
extern Boolean featureSet3_1;			// 3.1 ROM or later
extern Boolean featureSet3_2;			// 3.2 ROM or later
extern Boolean featureSet3_5;			// 3.5 ROM or later

extern Boolean colourDevice;			// Does device support colour ?
extern Boolean useColour;				// Use colour where appropriate

//	History.c

extern int historyMonths;				// Months of history to display
extern UInt32 historyGoTo;				// Date to show as end of history if nonzero
extern void historyLoad(FormPtr f);

//	Monthlog.c

extern UInt16 logGotoMonth;				// Month (database record number) to display
extern Boolean goToToday;				// Position log to edit today's entry
extern Boolean showToday;				// Position log to today's entry (but don't edit)
extern UInt16 showSpecificDay ;			// Initally make this day (1-31) visible if nonzero
extern rung recordRung;					// Record rung for today
extern void editWeight(char *txt, weight w);
extern void monthLogShowToday(void);	// Display today's entry in the Monthly Log form
extern void monthLogLoad(FormPtr f);

//	Monthlst.c

extern void monthListLoad(FormPtr f);

//	Prefs.c

#ifdef OS20BUG_WORKAROUND
#define NumIn(s)	ourStrDelocalizeNumber(s, numberThousands, numberDecimal)
#else
#define NumIn(s)	StrDelocalizeNumber(s, numberThousands, numberDecimal)
#endif
#define NumOut(s)	StrLocalizeNumber(s, numberThousands, numberDecimal)
#define PrefFlag(f)	((prefs.preFlags & (1 << (f))) ? true : false)
#define PrefFlagSet(f, b) prefs.preFlags = (prefs.preFlags & (~(1 << (f)))) | ((b) ? (1 << (f)) : 0)
#define PrefColourIndex(col) prefs.colourScheme.col.index
extern Char numberThousands, numberDecimal;	// Thousands and decimal separators for locale
extern Boolean hidePrivateData;			// Hide private records in database ?					
extern struct hdPreferences prefs;		// Application preferences
extern void getPreferences(void);
extern void savePreferences(void);
extern void preferencesLoad(FormPtr f);

//	Trendan.c

extern void editDeltaWeight(char *deltaw, char *sgn, signed long whole, signed long frac);
extern void trendanLoad(FormPtr f);

//	Trendfit.c

struct iTrendFitContext {	// Trend fitting context for fixed-point integer computation
	int n;
	double s1;
	long s2, s3, s4;
};
extern double fitSlopeToTrend(struct hdMonth *m, UInt16 firstDay, UInt16 lastDay);
extern Boolean fitSlopeToMultiMonthTrend(UInt16 sYear, UInt16 sMonth, UInt16 sDay,
										 UInt16 eYear, UInt16 eMonth, UInt16 eDay,
										 double *trend);
#ifdef NEED_DFIT_MULTI
extern Boolean DfitSlopeToMultiMonthTrend(UInt16 sYear, UInt16 sMonth, UInt16 sDay,
										  UInt16 eYear, UInt16 eMonth, UInt16 eDay,
										  double *slope);
#endif
//	Low-level routines
extern void fitSlopeStart(struct iTrendFitContext *c);
extern void fitSlopePoint(struct iTrendFitContext *c, trend t);
extern double fitSlopeEnd(struct iTrendFitContext *c);
extern void DfitSlopeStart(void);
extern void DfitSlopePoint(trend t);
extern double DfitSlopeEnd(void);

//	Units.c

/*	The "cvt" macros avoid the function call overhead for the most
	common case in which the from and to units are the same.
	Use them when bulk processing a number of items when you expect
	the storage and display units are identical.  */
	
#define cvtWeight(w, from, to) (((from) == (to)) ? (w) : convertWeight((w), (from), (to))) 
extern weight convertWeight(weight w, weightunit from, weightunit to);
#define cvtTrend(t, from, to) (((from) == (to)) ? (t) : convertTrend((t), (from), (to))) 
extern trend convertTrend(trend t, weightunit from, weightunit to);
#define cvtLong(l, from, to) (((from) == (to)) ? (l) : convertLong((l), (from), (to)))
extern signed long convertLong(signed long l, weightunit from, weightunit to);

//	Utility.c

extern void stringListItem(Int16 rsrc, UInt16 which, char *buffer, UInt16 buflen);
extern void stringResource(Int16 rsrc, char *buffer, UInt16 buflen);
#ifdef NEEDED
extern char *stringDup(char *s);
#endif
#ifdef OS20BUG_WORKAROUND
extern Char *ourStrDelocalizeNumber(Char *s, Char thousandSeparator, Char decimalSeparator);
#endif
extern void fldSetText(FieldPtr fp, char *text);
extern void fldSetNumber(FieldPtr fp, int value);
extern void drawInField(char *s, RectanglePtr extents);
extern void drawInFieldInverted(char *s, RectanglePtr extents);
extern void drawInFieldRight(char *s, RectanglePtr extents);
extern void drawInFieldCentre(char *s, RectanglePtr extents);

//	Yearlist.c

extern void yearListLoad(FormPtr f);

//	Yearview.c

extern int yearGotoYear;					// Which year to show (current year if -1)
extern void yearViewLoad(FormPtr f);
