/*
	
					The Hacker's Diet

					  Trend Analysis
					 
		This file manages the historical trend analysis form.
		
*/

#include "Hackdiet.h"

#define SecondsPerDay	(24L * 60 * 60)
#define DaysToSeconds(d) ((d) * SecondsPerDay)
#define SecondsToDays(s) ((s) / SecondsPerDay)

#define TrendIntervals		6		// Number of trend intervals

static struct {
	UInt32 firstDate;				// First day of interval
	UInt32 lastDate;				// Last day of interval
	double tslope;					// Evaluated trend slope
	struct iTrendFitContext ctx;	// Trend fitting computation context
} trendIntervalContext[TrendIntervals];

static Int8 trendIntervalLength[TrendIntervals] = {
	7,								// Week
	14,								// Fortnight
	-1,								// Month
	-3,								// Quarter
	-6,								// Six months
	-12								// Year
};

//	editDeltaWeight  --  Edit weight change from sign, whole, and decimal components

void editDeltaWeight(char *deltaw, char *sgn, signed long whole, signed long frac)
{
#ifdef OS20BUG_WORKAROUND
		char sfrac[4];

		/*	PalmOS 2.0 doesn't support right justification or zero filling
			of number fields in StrPrintF by format phrases such as "%02ld".
			We work around this by editing the fractional part to a
			left-justified string, then right justify and zero fill by hand.  */
			
		StrPrintF(sfrac, "%ld", frac);
		if (sfrac[1] == 0) {
			sfrac[1] = sfrac[0];
			sfrac[0] = '0';
			sfrac[2] = 0;
		}
		StrPrintF(deltaw, "%s%ld.%s", sgn, whole, sfrac);
#else
		StrPrintF(deltaw, "%s%ld.%02ld", sgn, whole, frac);
#endif
}

/*	adjustDateByInterval  --  Determine the date (seconds) resulting from
							  backing up from startDate (DateTimeType) by
							  intervalLength days (if positive) or
							  months (if negative).  */

static UInt32 adjustDateByInterval(DateTimeType *startDate, int intervalLength)
{
	DateTimeType strend;

	/*	intervalLength is positive to indicate a number of days and negative
		to denote a number of months.  If a number of months is given, we back
		up to the corresponding day the given number of months earlier, then
		adjust the day to the end of the month if there are fewer days in the
		start month than the present day in the end month.  */
	
	strend = *startDate;
	if (intervalLength > 0) {
		TimAdjust(&strend, -(intervalLength * (24L * 60 * 60)));
	} else {
		while (intervalLength < 0) {
			if (strend.month == 1) {
				strend.month = 12;
				strend.year--;
			} else {
				strend.month--;
			}
			intervalLength++;
		}
		if (strend.day > DaysInMonth(strend.month, strend.year)) {
			strend.day = DaysInMonth(strend.month, strend.year);
		}
	}
	return TimDateTimeToSeconds(&strend);
}

//	drawAnalysisItem  --  Paint a row of of the trend analysis

static void drawAnalysisItem(FormPtr f, UInt16 rowLabel,
							 UInt16 weightX, UInt16 calX,
							 Boolean tok, double tslope)
{
	signed long weeklyTrendChange, whole, frac;
	char deltaw[24], deltacal[24];
	char *sgn;
	long dcal;
	Int16 itemY;
	RectangleType r;

	FrmGetObjectBounds(f, FrmGetObjectIndex(f, rowLabel), &r);
	itemY = r.topLeft.y;
	
	if (tok) {
		dcal = (long) (EnergyPerWeightUnit(prefs.displayUnit) * tslope);
		weeklyTrendChange = (signed long) (tslope * 1000000 * 7);
		weeklyTrendChange = weeklyTrendChange;
	
		sgn = (weeklyTrendChange < 0) ? "-" : "+";
		if (weeklyTrendChange < 0) {
			weeklyTrendChange = -weeklyTrendChange;
		}
		weeklyTrendChange += 5000L;
		whole = weeklyTrendChange  / 1000000L;
		frac = (weeklyTrendChange / 10000L) % 100;
		if (whole == 0 && frac == 0) {
			sgn = "";
		}
		editDeltaWeight(deltaw, sgn, whole, frac);
		NumOut(deltaw);
		StrPrintF(deltacal, "%s%ld", (dcal == 0) ? "" : (dcal > 0 ? "+" : "-"), abs(dcal));
	
		RctSetRectangle(&r, weightX, itemY, 36, 15);
		drawInFieldRight(deltaw, &r);
		
		RctSetRectangle(&r, calX, itemY, 30, 15);
		drawInFieldRight(deltacal, &r);
	} else {
	
		/*	If there is insufficient data in the database to compute a
			trend of this length, blank its label.  */

		RctSetRectangle(&r, 0, itemY, weightX, 15);
		WinEraseRectangle(&r, 0);
	}

}

//	paintAnalysis  --  Paint trend analysis form

static void paintAnalysis(void)
{
	FormPtr f = FrmGetActiveForm();
	Int16 weightX, calX;
	DateTimeType today, firstMonth, originDate;
	int i, j, iter, dbindex;
	UInt32 now, then, origin, currentDate;
	UInt16 iyear, imonth, nmonths, lyear, lmonth;
	trend ltrend = 0, dtrend;
	struct hdMonth *m;
	UInt32 mstart;
	RectangleType fBounds;

	TimSecondsToDateTime(TimGetSeconds(), &today);
	today.hour = today.minute = today.second = 0;
	now = TimDateTimeToSeconds(&today);
	
	//	Get X offset of columns where we will paint the trend numbers

	FrmGetObjectBounds(f, FrmGetObjectIndex(f, TrendAnalysisWeightTrendLabel), &fBounds);
	weightX = fBounds.topLeft.x;
	FrmGetObjectBounds(f, FrmGetObjectIndex(f, TrendAnalysisCalorieTrendLabel), &fBounds);
	calX = fBounds.topLeft.x;

	for (i = 0; i < TrendIntervals; i++) {
		trendIntervalContext[i].firstDate = adjustDateByInterval(&today, trendIntervalLength[i]);
		trendIntervalContext[i].lastDate = now;
		fitSlopeStart(&trendIntervalContext[i].ctx);
	}
	
	//	Peek at the first month in the database to obtain its date
	
	iter = -1;
	hdIterateMonths(&iter, &iyear, &imonth);
	firstMonth = today;						// Cheap way to initialise time fields
	firstMonth.year = iyear;
	firstMonth.month = imonth + 1;
	firstMonth.day = 1;
	then = TimDateTimeToSeconds(&firstMonth);
	iter = -1;
	hdIterateMonths(&iter, &iyear, &imonth);
	
	/*	Riffle through the intervals and zap any whose start date
		precedes the first entry in the database.  In the process,
		we obtain the origin--the first month containing data
		relevant to the trend evaluation.  */
		
	for (i = 0; i < TrendIntervals; i++) {
		if (trendIntervalContext[i].firstDate < then) {
			trendIntervalContext[i].lastDate = 0;
		} else {
			origin = trendIntervalContext[i].firstDate;
		}
	}
	
	/*	Now we're ready to pass the months which span the longest
		of the surviving intervals, adding to the trend computation
		all dates within their calculated ranges.  */
	
	TimSecondsToDateTime(origin, &originDate);
	originDate.day = 1;
	origin = TimDateTimeToSeconds(&originDate);	
	dbindex = hdFindMonth(originDate.year, originDate.month - 1);
	
	/*	If there are entire months for which the database contains no
		data, we may have come up empty  looking for the first putative
		month in the interval.  If this happens, we walk through the
		database looking for the last month prior to the start of the
		interval and use that as a proxy for the missing month.  */
		
	if (dbindex == -1) {
		UInt16 liter = -1;
		while (true) {
			hdIterateMonths(&iter, &iyear, &imonth);
			if (iter == -1) {
				break;
			}
			if ((iyear > originDate.year) ||
				((iyear == originDate.year) && ((imonth + 1) > originDate.month))) {
				iter = -1;
				hdIterateMonths(&iter, &iyear, &imonth);
				break;
			}
			liter = iter;
		}
		dbindex = liter;
	}
	
	nmonths = hdNumMonths();
	lyear = lmonth = 0;
	currentDate = origin;
	while (dbindex < nmonths) {
	
		m = hdGetMonthUnpacked(dbindex);
		if (m == NULL) {
			break;
		}
		
		/*	If one or more months are missing from the database, fake
			them by filling in the last trend value we encountered.  */
		
		if (lyear != 0) {
			while (true) {
				UInt16 dim;
				
				lmonth++;
				if (lmonth >= 12) {
					lyear++;
					lmonth = 0;
				}
				if ((lyear == m->year) && (lmonth == m->month)) {
					break;
				}
				dim = DaysInMonth(lmonth + 1, lyear); 
				for (j = 0; j < dim; j++) {
					for (i = 0; i < TrendIntervals; i++) {
						if (trendIntervalContext[i].lastDate != 0) {
							if ((currentDate >= trendIntervalContext[i].firstDate) &&
								(currentDate <= trendIntervalContext[i].lastDate)) {
								fitSlopePoint(&trendIntervalContext[i].ctx, ltrend);
							}
						}
					}
					currentDate += DaysToSeconds(1);
					if (currentDate > now) {
						goto fakebail;
					}
				}
			}
		}
fakebail:;

		firstMonth.year = m->year;
		firstMonth.month = m->month + 1;
		firstMonth.day = 1;
		mstart = TimDateTimeToSeconds(&firstMonth);
		
		//	If the start of this month is after the end of the interval, bail out
		
		if (mstart > now) {
			hdDisposeMonthUnpacked(m);
			break;
		}
		
		/*	Now we're finally ready to walk through the days of the
			month and include them in the trend computation.  */
			
		for (j = 0; j < m->days; j++) {
			dtrend = cvtTrend(m->trendArray[j], (weightunit) m->weightUnit, prefs.displayUnit);
			for (i = 0; i < TrendIntervals; i++) {
				if (trendIntervalContext[i].lastDate != 0) {
					if ((currentDate >= trendIntervalContext[i].firstDate) &&
						(currentDate <= trendIntervalContext[i].lastDate)) {
						fitSlopePoint(&trendIntervalContext[i].ctx, dtrend);
					}
				}
			}
			currentDate += DaysToSeconds(1);
			if (currentDate > now) {
				break;
			}
		}
		ltrend = dtrend;
		
		lyear = m->year;
		lmonth = m->month;
		hdDisposeMonthUnpacked(m);
		dbindex++;
		if (currentDate > now) {
			break;
		}
	}
		
	//	Complete trend calculation for each interval
	
	for (i = 0; i < TrendIntervals; i++) {
		if (trendIntervalContext[i].lastDate != 0) {
			trendIntervalContext[i].tslope = fitSlopeEnd(&trendIntervalContext[i].ctx);
		}
	}
	
	drawAnalysisItem(f, TrendAnalysisLastWeekLabel, weightX, calX,
						trendIntervalContext[0].lastDate != 0 && trendIntervalContext[0].ctx.n > 0, trendIntervalContext[0].tslope);
	drawAnalysisItem(f, TrendAnalysisLast2WeeksLabel, weightX, calX,
						trendIntervalContext[1].lastDate != 0 && trendIntervalContext[1].ctx.n > 0, trendIntervalContext[1].tslope);
	drawAnalysisItem(f, TrendAnalysisLastMonthLabel, weightX, calX,
						trendIntervalContext[2].lastDate != 0 && trendIntervalContext[2].ctx.n > 0, trendIntervalContext[2].tslope);
	drawAnalysisItem(f, TrendAnalysisLastQuarterLabel, weightX, calX,
						trendIntervalContext[3].lastDate != 0 && trendIntervalContext[3].ctx.n > 0, trendIntervalContext[3].tslope);
	drawAnalysisItem(f, TrendAnalysisLastSixMonthsLabel, weightX, calX,
						trendIntervalContext[4].lastDate != 0 && trendIntervalContext[4].ctx.n > 0, trendIntervalContext[4].tslope);
	drawAnalysisItem(f, TrendAnalysisLastYearLabel, weightX, calX,
						trendIntervalContext[5].lastDate != 0 && trendIntervalContext[5].ctx.n > 0, trendIntervalContext[5].tslope);
}

/*	trendanHandleEvent  --  Handle event while the trend analysis form
							is displayed.  */
							  
static Boolean trendanHandleEvent(EventPtr event)
{
	Boolean handled;
#ifdef __GNUC__
	CALLBACK_PROLOGUE
#endif
	handled = false;


	switch (event->eType) {
	  	
  		case frmOpenEvent:
  			{
				static char ltext[60], energyColumnLabel[12];
				char wunits[16], fmt[60];
#ifdef OLDWAY
				ControlPtr lp;
#endif
				FormPtr f = FrmGetActiveForm();
				weightunit wdu = prefs.displayUnit;
			
				//	Set label of weight trend to selected weight unit
			
				if (wdu == wStone) {
					wdu = wPound;				// Show trend in pounds for log in stones
				}	
				stringListItem(WeightUnitAbbreviationsStringList, wdu, wunits, sizeof wunits);
				stringResource(TrendUnitsPerWeekFormatString, fmt, sizeof fmt);
				StrPrintF(ltext, fmt, wunits);
#ifdef OLDWAY
				lp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, TrendAnalysisWeightTrendLabel));
				CtlSetLabel(lp, ltext);
#else
				FrmCopyLabel(f, TrendAnalysisWeightTrendLabel, ltext);
#endif
				if (PrefFlag(pfEnergyJoule)) {
					stringResource(KilojoulesPerDayString, energyColumnLabel, sizeof energyColumnLabel);
#ifdef OLDWAY
					lp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, TrendAnalysisCalorieTrendLabel));
					CtlSetLabel(lp, energyColumnLabel);
#else
					FrmCopyLabel(f, TrendAnalysisCalorieTrendLabel, energyColumnLabel);
#endif
				}
	  			FrmDrawForm(f);					// Paint to show titles before we calculate	
				paintAnalysis();
				handled = true;
			}
			break;
			
		case frmCloseEvent:
			FrmEraseForm(FrmGetActiveForm());
			FrmDeleteForm(FrmGetActiveForm());
			handled = true;
			break;
						
		case ctlSelectEvent:
			switch(event->data.ctlSelect.controlID) {
				case TrendAnalysisDoneButton:			// Done button: Back to today's log
					monthLogShowToday();
					handled = true;
					break;
					
				case TrendAnalysisTodayButton:			// Today button: Show log for today
					{
						DateTimeType dt;
	
						TimSecondsToDateTime(TimGetSeconds(), &dt);
						goToToday = true;
						hdEnsureMonthPresent(dt.year, dt.month - 1);
						logGotoMonth = hdFindMonth(dt.year, dt.month - 1);
						FrmGotoForm(MonthlyLogForm);
						handled = true;
					}
					break;
			}
			break;
	}
#ifdef __GNUC__
	CALLBACK_EPILOGUE
#endif
	return handled;
}

//	trendanLoad  --  Initialise trend analysis form. */

void trendanLoad(FormPtr f)
{
	if (hidePrivateData && prefs.privateLogs) {
		FrmAlert(PrivateLogAlert);
		FrmGotoForm(PreferencesForm);
	}		
	FrmSetEventHandler(f, trendanHandleEvent);	
}