/*
	
					The Hacker's Diet

						Utilities
					 
*/

#include "Hackdiet.h"			

/*	stringListItem  --  Given the resource ID of a string list
						resource and an item number within it, copy
						that string to a buffer provided by the caller.
						The string will be truncated if necessary to
						avoid overflowing the caller's buffer.  If
						there are fewer strings in the string list
						than that requested or the string list resource
						doesn't exist, the caller receives the null
						string.  If the string list specifies a prefix, it
						will be prefixed to the string the caller receives.  */

#define StringListResourceType	'tSTL'
						
void stringListItem(Int16 rsrc, UInt16 which, char *buffer, UInt16 buflen)
{
	MemHandle stl;
	char *ip, *cp;
	UInt16 n, nstrings;

	assert(buflen > 0);
	*buffer = 0;
	cp = buffer;
	stl = DmGetResource(StringListResourceType, rsrc);
	if (stl != NULL) {
		ip = MemHandleLock(stl);
		
		if (StrLen(ip) > 0) {
			if (buflen < (StrLen(ip) + 1)) {
				MemMove(cp, ip, buflen - 1);
				cp[buflen - 1] = 0;
				goto bailout;			// Prefix alone filled buffer
			} else {
				StrCopy(cp, ip);
				cp += StrLen(ip);
				buflen -= StrLen(ip);
			}
		}
		ip += StrLen(ip) + 1;			// Advance past prefix, if any
		MemMove(&nstrings, ip, 2);		// Sidestep any possible alignment problems
		ip += 2;
		
		if (which >= nstrings) {		// Does string index exceed strings in resource ?
			*buffer = 0;				// Wipe out any prefix
			goto bailout;				// Return null string
		}
		
		n = 0;
		while (true) {
			assert(n < nstrings);
			if (which == n) {
				if (buflen < (StrLen(ip) + 1)) {
					MemMove(cp, ip, buflen - 1);
					cp[buflen - 1] = 0;
				} else {
					StrCopy(cp, ip);
					cp += StrLen(ip);
					buflen -= StrLen(ip);
				}
				break;
			}
			ip += StrLen(ip) + 1;
			n++;
		}
bailout:;		
		MemHandleUnlock(stl);
		DmReleaseResource(stl);
	}
}

//	 stringResource  --  Get text of a string resource

#define StringResourceType	'tSTR'
						
void stringResource(Int16 rsrc, char *buffer, UInt16 buflen)
{
	MemHandle str;
	char *ip, *cp;

	assert(buflen > 0);
	*buffer = 0;
	cp = buffer;
	str = DmGetResource(StringResourceType, rsrc);
	if (str != NULL) {
		ip = MemHandleLock(str);
		
		
		if (buflen < (StrLen(ip) + 1)) {
			MemMove(cp, ip, buflen - 1);
			cp[buflen - 1] = 0;
		} else {
			StrCopy(cp, ip);
			cp += StrLen(ip);
			buflen -= StrLen(ip);
		}
		
		MemHandleUnlock(str);
		DmReleaseResource(str);
	}
}

#ifdef NEEDED

/*	stringDup  --  Return a dynamically allocated copy of a string.

				   CAUTION!!!  This function may cause the heap to be
				   			   compacted, which can change form and form
				   			   object pointers and any other non-locked
				   			   memory.  Be sure to recover such pointers
				   			   after calling stringDup() or the gremlins
				   			   are gonna get ya.  */

char *stringDup(char *s)
{
	char *d = MemPtrNew(StrLen(s) + 1);
	
	if (d != NULL) {
		StrCopy(d, s);
	}
	return d;
}
#endif

#ifdef OS20BUG_WORKAROUND

/*	ourStrDelocalizeNumber  --  Local re-implementation of PalmOS StrDelocalizeNumber()
								which doesn't work on pre-3.0 ROMs.  */

Char *ourStrDelocalizeNumber(Char *s, Char thousandSeparator, Char decimalSeparator)
{
	Char *n = s;
	
	if (featureSet3_0) {
		return StrDelocalizeNumber(s, thousandSeparator, decimalSeparator);
	} else {
		while (*n) {
			if (*n == thousandSeparator) {
				*n = ',';
			} else if (*n == decimalSeparator) {
				*n = '.';
			}
			n++;
		}
	}
	return s;
}
#endif

//	fldSetText  --  Set text of field from in-memory string

void fldSetText(FieldPtr fp, char *text)
{
	MemHandle otext = FldGetTextHandle(fp);
	MemHandle ntext = MemHandleNew(StrLen(text) + 1);
	char *cp;
	
	if (ntext != NULL) {
		cp = MemHandleLock(ntext);
		StrCopy(cp, text);
		MemHandleUnlock(ntext);
		FldSetTextHandle(fp, ntext);
		FldDrawField(fp);
		if (otext != NULL) {
			MemHandleFree(otext);
		}
	}
}

//	fldSetNumber  --  Set a field's text to a decimal number

void fldSetNumber(FieldPtr fp, int value)
{
	char n[12];
	
	StrPrintF(n, "%d", value);
	fldSetText(fp, n);
}

//	drawInField  --  Paint text in rectangle, but constrain to its bounds

void drawInField(char *s, RectanglePtr extents)
{
	Int16 l = StrLen(s),
		  w = extents->extent.x;
	Boolean truncated;
	
	FntCharsInWidth(s, &w, &l, &truncated);
	WinDrawChars(s, l, extents->topLeft.x, extents->topLeft.y);	
}

//	drawInFieldInverted  --  Paint inverted text in rectangle, but constrain to its bounds

void drawInFieldInverted(char *s, RectanglePtr extents)
{
	Int16 l = StrLen(s),
		  w = extents->extent.x;
	Boolean truncated;
	
	FntCharsInWidth(s, &w, &l, &truncated);
	WinDrawInvertedChars(s, l, extents->topLeft.x, extents->topLeft.y);	
}

/*	drawInFieldRight  --  Paint text right-justified in rectangle,
						  constrained to its bounds.  */
						  
void drawInFieldRight(char *s, RectanglePtr extents)
{
	Int16 l = StrLen(s),
		  w = extents->extent.x;
	Boolean truncated;
	
	FntCharsInWidth(s, &w, &l, &truncated);
	WinDrawChars(s, l, extents->topLeft.x + (extents->extent.x - w), extents->topLeft.y);
}

/*	drawInFieldCentre  --  Paint text centred in rectangle,
						   constrained to its bounds.  */
						  
void drawInFieldCentre(char *s, RectanglePtr extents)
{
	Int16 l = StrLen(s),
		  w = extents->extent.x;
	Boolean truncated;
	
	FntCharsInWidth(s, &w, &l, &truncated);
	WinDrawChars(s, l, extents->topLeft.x + (extents->extent.x - w) / 2, extents->topLeft.y);
}
